{-# LANGUAGE DeriveFunctor, RankNTypes, FlexibleInstances #-}

module Main where

import Control.Monad
import Criterion.Main

import FreeMonadPlus
import CayleyMonadPlus

newtype Identity a = Id {runId :: a} deriving Functor

class MonadPlus m => Bunch m where
  wrap :: m a -> m a

instance Bunch (FMP Identity) where
  wrap xf = FMP [ConFMP (Id xf)]

forest :: MonadPlus m => (forall a. Identity a -> m a) -> Int -> m ()
forest _ 0     =  return ()
forest wrap n  =  (forest wrap (n-1) >>= (wrap . Id)) `mplus` (wrap . Id) ()

solutionCount :: FMP Identity () -> Int
solutionCount (FMP f)  =  sum (map go f)
  where  go (PureFMP _)  =  1
         go (ConFMP (Id f))  =  solutionCount f

bench1group n = bgroup ("leftmplus" ++ "/" ++ show n)
  [ bench (show i) $ nf (solutionCount . forest (wrap . return . runId)) (n * i)
  | i <- [1,2,4,8,16]
  ]

bench2group n = bgroup ("leftmplus - Cayley" ++ "/" ++ show n)
  [ bench (show i) $ nf (solutionCount . down . forest (up . (wrap . return . runId))) (n * i)
  | i <- [1,2,4,8,16,32] 
  ]

main = defaultMain [
  bench1group 64,
  bench2group 64
  ]
