{-# LANGUAGE RankNTypes, TypeOperators, ExistentialQuantification #-}

-- Section 6.5

module CayleyAlternative where

import Control.Applicative
import Control.Monad
import Prelude

-- Day convolution

data (:*:) f g a = forall b. Day (f (b -> a)) (g b)

-- Exponential for convolution

data (:**>) f g x = ED { unED :: forall y . f y -> g (x, y) }

instance Functor g => Functor (f :**> g) where
  fmap f (ED g) = ED (\y -> fmap (\(a,b) -> (f a, b)) (g y))

ladj3  ::  (Functor f, Functor g, Functor h) =>
           (forall x. (f :*: g) x -> h x) -> (forall x. f x -> (g :**> h) x) 
ladj3 f = \fx -> ED (\gy -> f (Day (fmap (,) fx) gy))

radj3  ::  (Functor f, Functor g, Functor h) =>
           (forall x. f x -> (g :**> h) x) -> (forall x. (f :*: g) x -> h x)
radj3 f  =  \(Day ff gy) -> fmap  (uncurry ($)) (unED (f ff) gy)

-- Exponential for product

newtype  (:*>) f g x = Exp { unExp:: forall y . (x -> y) -> (f y -> g y) }

instance Functor (f :*> g) where
  fmap f m = Exp (\k -> unExp m (k . f))

ladj2  ::  (Functor f, Functor g, Functor h) =>
           (forall x. (f x, g x) -> h x) -> (forall x. f x -> (g :*> h) x)
ladj2 f  =  \fx -> Exp (\t -> \gy -> f (fmap t fx,gy))

radj2  ::  (Functor f, Functor g, Functor h) =>
           (forall x. f x -> (g :*> h) x) -> (forall x. (f x, g x) -> h x)
radj2 f = \fgx -> unExp (f (fst fgx)) id (snd fgx)

-- Double Cayley representation

newtype Opt f x =  Opt { unOpt :: ((f :*> f) :**> (f :*> f)) x }

instance Functor f => Functor (Opt f) where
  fmap f (Opt z) = Opt (ED (\fx ->
    fmap (\(x,y) -> (f x, y)) (unED z fx)))

instance Functor f => Applicative (Opt f) where
  pure v = Opt (ED (\f -> fmap (\y -> (v, y)) f))
  (Opt (ED h)) <*> (Opt (ED v)) = fmap (uncurry ($))
    (Opt (ED (\f -> fmap (\(b, (a, y)) -> ((a, b), y)) (v (h f)))))

instance Functor f => Alternative (Opt f) where
  empty = Opt (ED (const (Exp (\h x -> x))))
  Opt f <|> Opt g = Opt (ED (\h ->
    Exp (\j i -> unExp (unED f h) j (unExp (unED g h) j i))))

up :: Alternative f => f a -> Opt f a
up x = Opt (ED  (\g -> Exp (\f fx ->
                unExp g (flip (curry f)) empty <*> x  <|>  fx)))

down :: Alternative f => Opt f a -> f a
down (Opt (ED f)) =  unExp (f (Exp (\g i ->  pure (g ()) <|> i))) fst empty
