{-# LANGUAGE RankNTypes, TypeOperators #-}

-- Section 6.4

module CayleyMonadPlus where

import Control.Monad
import Prelude

-- Exponential for composition

newtype (:.>)  f g x = Ran { unRan :: forall y . (x -> f y) -> g y }

instance Functor (g :.> h) where
  fmap f m = Ran (\k -> unRan m (k . f))

ladj1  ::  (Functor f, Functor g, Functor h) =>
           (forall x. f (g x) -> h x) -> (forall x. f x -> (g :.> h) x)
ladj1 f  =  \gx -> Ran (\k -> f (fmap k gx)) 

radj1  ::  (Functor f, Functor g, Functor h) =>
           (forall x. f x -> (g :.> h) x) -> (forall x. f (g x) -> h x)
radj1 f  =  \fgx -> unRan (f fgx) id

-- Exponential for product

newtype  (:*>) f g x = Exp { unExp:: forall y . (x -> y) -> (f y -> g y) }

instance Functor (f :*> g) where
  fmap f m = Exp (\k -> unExp m (k . f))

ladj2  ::  (Functor f, Functor g, Functor h) =>
           (forall x. (f x, g x) -> h x) -> (forall x. f x -> (g :*> h) x)
ladj2 f  =  \fx -> Exp (\t -> \gy -> f (fmap t fx,gy))

radj2  ::  (Functor f, Functor g, Functor h) =>
           (forall x. f x -> (g :*> h) x) -> (forall x. (f x, g x) -> h x)
radj2 f = \fgx -> unExp (f (fst fgx)) id (snd fgx)

-- Double Cayley representation

newtype Opt f x =  Opt { unOpt :: ((f :*> f) :.> (f :*> f)) x }

proj (Opt (Ran a)) = a

instance Monad (Opt f) where
  return x = Opt (Ran (\f -> f x))
  Opt (Ran m) >>= f = Opt (Ran (\g -> m (\a -> proj (f a) g)))

instance MonadPlus (Opt f) where
  mzero = Opt (Ran (\k -> Exp (\c x -> x)))
  mplus m n = Opt (Ran (\sk ->
    Exp (\f fk -> unExp (proj m sk) f (unExp (proj n sk) f fk))))

up :: Monad m => m a -> Opt m a
up x = Opt  (Ran (\g ->  Exp  (\h m -> x >>= \a -> unExp (g a) h m)))

down :: MonadPlus m => Opt m a -> m a
down (Opt (Ran f)) =  unExp  (f  (\x -> Exp  (\h m -> 
                                                return (h x) `mplus` m)))
                             id mzero
