{-# LANGUAGE TypeOperators #-}

-- Section 4

module Examples where

import Control.Monad
import Control.Applicative hiding (ZipList)

-- ZipList

newtype ZipList a = ZL [a]

instance Applicative ZipList where
    pure x           = ZL (repeat x)
    ZL fs <*> ZL xs  = ZL (zipWith ($) fs xs)

instance Functor ZipList where
    fmap f (ZL xs) = ZL (map f xs)

instance Alternative ZipList where
  empty            = ZL []
  ZL []      <|> ZL ys      = ZL ys
  ZL xs      <|> ZL []      = ZL xs
  ZL (x:xs)  <|> ZL (y:ys)  = ZL (x : rs)
    where ZL rs = ZL xs <|> ZL ys

-- Parser
newtype Parser m a = Parser { unParser :: String -> m (a, String) }

instance Monad m => Monad (Parser m) where
  return x  = Parser (\s -> return (x,s))
  x >>= f   = Parser (\s -> unParser x s >>= \(y, s') -> unParser (f y) s')

instance MonadPlus m => MonadPlus (Parser m) where
  mzero        = Parser (\s -> mzero)
  x `mplus` y  =  Parser (\s -> unParser x s `mplus` unParser y s)

instance Monad m => Functor (Parser m) where
  fmap f x = Parser (\s -> unParser x s >>= \(a, s') -> return (f a, s'))

instance Monad m => Applicative (Parser m) where
  pure x = return x
  f <*> x = ap f x

instance MonadPlus m => Alternative (Parser m) where
  empty = mzero
  x <|> y = x `mplus` y


-- Parser Maybe is not a non-determinism applicative


char :: Char -> Parser Maybe ()
char d = Parser (\s -> case  s of
                             ""      ->   Nothing
                             (c:s')  ->   if c == d   then  Just ((), s')
                                                      else  Nothing)

p1 = (char 'a' <|> pure ()) >*< char 'a'

p2 = (char 'a' >*< char 'a') <|> (pure () >*< char 'a')

(>*<) :: Applicative p => p a -> p b -> p ()
p >*< q  =  pure (\x y -> ()) <*> p <*> q

-- Composition

newtype (f :.: g) x = Comp (f (g x))

instance (Functor f, Functor g) => Functor (f :.: g) where
   fmap f (Comp fgx) = Comp (fmap (fmap f) fgx) 

instance  (Applicative f,  Applicative g) => Applicative (f :.: g) where
   pure x  = Comp (pure (pure x))
   Comp fs <*> Comp xs = Comp (pure (<*>) <*> fs <*> xs)

instance  (Alternative f, Applicative g) => Alternative (f :.: g) where
   empty = Comp empty
   Comp xs <|> Comp ys = Comp (xs <|> ys)  
