{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE ExistentialQuantification #-}

-- Section 5.5

module FreeAlternative where

import Control.Monad
import Control.Applicative

import Prelude

data FreeAlt f a = FreeAlt { unFreeAlt :: [Alt f a] }
data Alt f a  =  PureFA a  |  forall b . ConFA (f (b -> a)) (FreeAlt f b)

instance Functor f => Functor (Alt f) where
  fmap f (PureFA a)   = PureFA (f a)
  fmap f (ConFA g x)  = ConFA (fmap (f .) g) x

instance Functor f => Functor (FreeAlt f) where
  fmap f (FreeAlt xs) = FreeAlt (fmap (fmap f) xs)

instance Functor f => Applicative (FreeAlt f) where
  pure x = FreeAlt [PureFA x]
  FreeAlt xs <*> v = FreeAlt (concatMap g xs)
    where  g (PureFA f)   = unFreeAlt (fmap f v)
           g (ConFA f c)  = [ConFA (fmap uncurry f) (pure (,) <*> c <*> v)]

instance Functor f => Alternative (FreeAlt f) where
  empty = FreeAlt []
  FreeAlt xs <|> FreeAlt ys = FreeAlt (xs ++ ys)

inj :: Functor f => f a -> FreeAlt f a
inj x = FreeAlt [ConFA (fmap (\z () -> z) x) (pure ())]

univ ::  Alternative g => (forall x. f x -> g x) -> FreeAlt f x -> g x
univ h (FreeAlt l)  = foldr (<|>) empty (map univT l)
                    where  univT (PureFA x)   = pure x
                           univT (ConFA q c)  = h q <*> univ h c
