{-# LANGUAGE ExistentialQuantification, TypeOperators, FlexibleInstances #-}

-- Section 7

module ListT where

import Control.Applicative
import Control.Monad
import Control.Monad.Trans

-- ListT monad transformer

newtype ListT m a = ListT { unListT :: m (Maybe (a, ListT m a)) }

instance Monad m => Monad (ListT m) where
  return x  =  ListT (return (Just (x,mzero)))   
  p >>= k   =  ListT (do  r <- unListT p
                          case r of
                            Nothing      ->  return Nothing
                            Just (x,p')  ->  unListT (mplus (k x) (p' >>= k)))

instance MonadTrans ListT where
  lift p  =  ListT (p >>= \x -> return (Just (x, mzero)))

instance Monad m => MonadPlus (ListT m) where
  mzero      =  ListT (return Nothing)
  mplus p q  =  ListT (do  r <- unListT p
                           case r of
                             Nothing      -> unListT q
                             Just (x,p')  -> return (Just (x,mplus p' q)))

runListT :: Monad m => ListT m a -> m [a]
runListT p  =  do  r <- unListT p
                   case r of
                     Nothing     -> return []
                     Just (x,p') -> runListT p' >>= return . (x:)

-- ListA applicative transformer

data ListA f a = forall b. f (b -> a) :+: Maybe (b, ListA f b)

instance Functor f => Functor (ListA f) where
  fmap f (op :+: r)  =  fmap (fmap f) op :+: r
instance Applicative f => Applicative (ListA f) where
  pure x  =  pure (const x) :+: Just ((), empty)
  (opx  :+: Nothing)      <*> ys               =  fmap uncurry opx :+: Nothing
  (opx  :+: Just (x,xs))  <*> ys@(opy :+: ry)  =  op :+: r
    where
      op  = ((\f g -> either (f x . g) (uncurry f)) <$> opx) <*> opy
      r   = fmap (\(y,ys') -> (Left y,  fmap Left ys' <|> 
                                        ((curry Right <$> xs) <*> ys))) ry

class ATrans t where
  alift :: Applicative f => f a -> t f a

instance ATrans ListA where
  alift p  =  fmap const p :+: Just ((),empty)

instance Applicative f => Alternative (ListA f) where
  empty    =  pure id :+: Nothing
  (op :+: Nothing)      <|>  (op' :+: r)  =  ((\x y -> y) <$> op <*> op') :+: r
  (op :+: Just (x,xs))  <|>  ys           =  
     fmap (flip either id) op :+: Just (Left x, fmap Left xs <|> fmap Right ys)

runListA :: Applicative f => ListA f a -> f [a]
runListA (op :+: Nothing)      =  const [] <$> op
runListA (op :+: Just (x,xs))  =  
                   (\f xs -> map f (x:xs)) <$> op <*> runListA xs

newtype (f :.: g) x = Comp (f (g x))

instance ATrans ((:.:) []) where
  alift p = Comp [p]

runL :: Applicative f => ([] :.: f) a -> f [a]
runL (Comp ps)  =  foldr (\x xs -> ((:) <$> x) <*> xs) (pure []) ps

dropL :: Applicative f => Int -> ([] :.: f) a -> f [a]
dropL n (Comp ps)  =  runL (Comp (drop n ps))

