{-# LANGUAGE DeriveFunctor, FlexibleInstances #-}

-- Section 8

module UnitalMonad where

import Control.Applicative
import Control.Monad

-- Unital monads

data IM m0 a  =  Pure a  |  Impure (m0 a)

class Functor m0 => Ideal m0 where
  join0 :: m0 (IM m0 a) -> IM m0 a

instance Ideal m0 => Monad (IM m0) where
  return x          =  Pure x
  Pure x     >>= k  =  k x
  Impure op  >>= k  =  join0 (fmap k op)

instance Ideal m0 => Functor (IM m0) where
  fmap = liftM
instance Ideal m0 => Applicative (IM m0) where
  pure   =  return
  (<*>)  =  ap
  
-- List monad as an unital monad

data List02 a = Cons0 | Cons2 a a [a] deriving Functor

fromL :: [a] -> IM List02 a
fromL []        =  Impure Cons0
fromL [x]       =  Pure x
fromL (x:y:xs)  =  Impure (Cons2 x y xs)

toL :: IM List02 a -> [a]
toL (Pure x)                 =  [x]
toL (Impure Cons0)           =  []
toL (Impure (Cons2 x y xs))  =  x:y:xs

instance Ideal List02 where
  join0 Cons0 = fromL []
  join0 (Cons2 x y xs) = fromL (concat (toL x : toL y : map toL xs))
  
instance MonadPlus (IM List02) where
  mzero = fromL []
  x `mplus` y = fromL (toL x ++ toL y)

-- Non-determinism monad from unital monad

data IMForest m0 a = InId a | InT1 (T1 m0 a) | InT2 (T2 m0 a)
data T1 m0 a = T1 (m0 (Either a (T2 m0 a)))
data T2 m0 a = T2 (List02 (Either a (T1 m0 a)))

instance Ideal m0 => Monad (IMForest m0) where
  return x = InId x
  InId v       >>= f  = f v
  InT1 (T1 v)  >>= f  = reB1 (join0 (fmap unB1 (fmap p1 v)))
    where  p1 (Left a)    = f a
           p1 (Right w)   = InT2 w >>= f
           unB1 (InId a)       = Pure (Left a)
           unB1 (InT2 v)       = Pure (Right v)
           unB1 (InT1 (T1 v))  = Impure v
           reB1 (Pure (Left a))    = InId a
           reB1 (Pure (Right v))   = InT2 v
           reB1 (Impure v)         = InT1 (T1 v)
  InT2 (T2 v) >>= f = reB2 (join0 (fmap unB2 (fmap p2 v)))
    where  p2 (Left a)   = f a
           p2 (Right w)  = InT1 w >>= f
           unB2 (InId a)       = Pure (Left a)
           unB2 (InT1 v)       = Pure (Right v)
           unB2 (InT2 (T2 v))  = Impure v
           reB2 (Pure (Left a))   = InId a
           reB2 (Pure (Right v))  = InT1 v
           reB2 (Impure v)        = InT2 (T2 v)

liftLeft :: Ideal m0 => IM m0 a -> IMForest m0 a
liftLeft (Pure x)     =  InId x
liftLeft (Impure m0)  =  InT1 (T1 (fmap Left m0))

liftRight :: Ideal m0 => IM List02 a -> IMForest m0 a
liftRight (Pure x)     =  InId x
liftRight (Impure m0)  =  InT2 (T2 (fmap Left m0))
 
instance Ideal m0 => MonadPlus (IMForest m0) where
  mzero        =  join  (liftRight mzero)
  x `mplus` y  =  join  (liftRight (return x `mplus` return y))
